/*
** Jo Sega Saturn Engine
** Copyright (c) 2012-2017, Johannes Fetz (johannesfetz@gmail.com)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Johannes Fetz nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL Johannes Fetz BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/***************************************************************************************************\
** Special Note : The 3D part on Jo Engine is still in development. So, some glitch may occur ;)
**                Btw, texture mapping for triangle base mesh (not quads) is experimental.
\***************************************************************************************************/

#include <jo/jo.h>
#include "ZT/ZT_COMMON.h"


/**Functions/variables added by XL2 **/
/**NOTE : Real-time gouraud shading - taken from Rockin'B demo**/

//GOURAUDTBL	grtbl_workArea5[750];
GOURAUDTBL		gourRealMax[GOUR_REAL_MAX];
Uint8			vwork[GOUR_REAL_MAX];
Uint8 enableRTG = 1;

/**Added by XL2 to use my own CD loading functions**/
#define     OPEN_MAX    (Sint32)5
#define     DIR_MAX     (Sint32)25
#define     RD_UNIT     (10)
#define     SECT_SIZE   (2048)
GfsDirTbl gfsDirTbl;
GfsDirName gfsDirName[DIR_MAX];
Uint32 gfsLibWork[GFS_WORK_SIZE(OPEN_MAX)/sizeof(Uint32)];
Sint32 gfsDirN;
void    ztCDinit(void)
{
    GFS_DIRTBL_TYPE(&gfsDirTbl) = GFS_DIR_NAME;
    GFS_DIRTBL_DIRNAME(&gfsDirTbl) = gfsDirName;
    GFS_DIRTBL_NDIR(&gfsDirTbl) = DIR_MAX;
    gfsDirN = GFS_Init(OPEN_MAX, gfsLibWork, &gfsDirTbl);
}
extern Sint8 SynchConst;
Sint32 framerate;
/****/

/**Taken from RB demo**/
static FIXED light[XYZ];
static ANGLE light_ang[XYZ];
void computeLight()
{
    FIXED light_init[XYZ] = { toFIXED(0.57735026), toFIXED(0.57735026), toFIXED(0.57735026) };

    slPushUnitMatrix();
    {

        slRotX(light_ang[X]);
        slRotY(light_ang[Y]);
        slRotZ(light_ang[Z]);
        //slTranslate(toFIXED(15.0), 0, 0);
        slCalcVector(light_init, light);
    }
    slPopMatrix();
    light_ang[X] += DEGtoANG(2.0);
    light_ang[Y] += DEGtoANG(2.0);
    light_ang[Z] += DEGtoANG(2.0);
}

/**Added by XL2 - Very basic function to display a 3D model **/
void display_model(entity_t * model, bool UseRealtimeGouraud)
{
    int i;
    for (i=0; i<model->nbMeshes; i++)
    {
       if (UseRealtimeGouraud) slPutPolygonX((XPDATA*)model->pol[i], light);
       else slPutPolygon((PDATA*)model->pol[i]);
    }
}
/**End of added functions and global variables**/




/**ADDED 2018-06-01 : ANIMATION EXAMPLE **/
typedef struct
{
    Uint8 currentFrm;
    Uint8 currentKeyFrm;
    Uint8 startFrm;
    Uint8 endFrm;
} animationControl;
animationControl AnimCtrlPlay1;
animationControl AnimCtrlPlay2;





/** WORK IN PROGRESS!!! **
* This function uses compressed vertex animation with linear interpolation to animate 3D models at low memory cost
* It currently only displays one mesh, just change this function if you want to display several models
* The normals compression is inspired by Quake 2, but I created my own 3D model with my own normals for the LUT
* The idea here is to create recreate the XPDATA by uncompressing the vertices/normals**/
#include "ZT/ANORM.h"
void display_animated_model(animationControl * animCtrl, entity_t * currentModel, bool UseRealtimeGouraud)
{
    XPDATA * currentPDATA = currentModel->pol[0];

    /**Sets the animation data**/
    animCtrl->currentFrm+=framerate;
    animCtrl->currentKeyFrm=animCtrl->currentFrm>>3;  //should be >>currentModel->AnimInterpolation; but I forgot to output 60 fps animations
    if (animCtrl->currentKeyFrm>=animCtrl->endFrm)    {
        animCtrl->currentFrm-=(animCtrl->endFrm-animCtrl->startFrm)<<3; //<<currentModel->AnimInterpolation;
        animCtrl->currentKeyFrm=animCtrl->currentFrm>>3;  //>>currentModel->AnimInterpolation;
    }

    Uint8 nextKeyFrm=animCtrl->currentKeyFrm+1;
    if (nextKeyFrm>=animCtrl->endFrm)
        nextKeyFrm=animCtrl->startFrm;

    compVert * curKeyFrame = (compVert*)currentModel->animation[animCtrl->currentKeyFrm]->cVert;
    compVert * nextKeyFrame = (compVert*)currentModel->animation[nextKeyFrm]->cVert;

    Uint16 interpFctr = animCtrl->currentFrm-(animCtrl->currentKeyFrm<<3);
    /*if (nextKeyFrm>animCtrl->currentKeyFrm) interpFctr = animCtrl->currentFrm-(animCtrl->currentKeyFrm<<3);
    else interpFctr = animCtrl->currentFrm-(animCtrl->currentKeyFrm<<3);*/

    /**Uncompress the vertices and apply linear interpolation**/
    register Uint32	i;
    Sint32 *dst=currentPDATA->pntbl[0];
    Sint16 *src=curKeyFrame[0];
    Sint16 *nxt=nextKeyFrame[0];
    for (i = 0; i < currentPDATA->nbPoint*sizeof(POINT); i+= sizeof(int)) {
		*dst++=(*src+(((*nxt-*src)*interpFctr)>>3))<<6;
		*src++; *nxt++;
    }
    *dst=currentPDATA->pltbl[0].norm[0];
    Uint8 *src2=currentModel->animation[animCtrl->currentKeyFrm]->cNorm;
    for (i = 0; i < currentPDATA->nbPolygon; i++)    {
    /**Not 100% sure which technique is faster**/
        /*currentPDATA->pltbl[i].norm[X]=anorms[*src2][X];
        currentPDATA->pltbl[i].norm[Y]=anorms[*src2][Y];
        currentPDATA->pltbl[i].norm[Z]=anorms[*src2++][Z];*/
        *dst++=anorms[*src2][X];
        *dst++=anorms[*src2][Y];
        *dst++=anorms[*src2++][Z];
        *dst++; *dst++;
    }
    if (UseRealtimeGouraud) slPutPolygonX(currentPDATA, light);
    else slPutPolygon((PDATA*)currentPDATA);
}




//You should just keep the animation for each bone, this is a bad example
ANGLE IdleAnim[][XYZ]=
{
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(0.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(1.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(3.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(4.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(6.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(7.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(6.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(4.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(3.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(1.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), -DEGtoANG(0.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), DEGtoANG(1.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), DEGtoANG(3.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), DEGtoANG(4.5)},
    {DEGtoANG(0.0), DEGtoANG(0.0), DEGtoANG(3.0)},
    {DEGtoANG(0.0), DEGtoANG(0.0), DEGtoANG(1.5)},
};

/**Bad animation example, but you should get the idea.**/
void animate_model(entity_t * currentModel, Uint32 key_frame, Uint32 anim_tim)
{
  slPushMatrix();
  {
        slRotX(IdleAnim[anim_tim][Z]);
        //BODY
        slPutPolygonX((XPDATA*)currentModel->pol[0], light);

        slPushMatrix();
        {
            //Head
            slPutPolygonX((XPDATA*)currentModel->pol[1], light);
        }
        slPopMatrix();

        slPushMatrix();
        {
            //Arms
            slRotX(-IdleAnim[anim_tim][Z]*4);
            slPutPolygonX((XPDATA*)currentModel->pol[2], light);
            slPutPolygonX((XPDATA*)currentModel->pol[4], light);
            slRotX(-IdleAnim[anim_tim][Z]*2);
            slPutPolygonX((XPDATA*)currentModel->pol[3], light);
            slPutPolygonX((XPDATA*)currentModel->pol[5], light);
        }
        slPopMatrix();

        if (currentModel->nbMeshes>10) //Tails' model
        {
            slPutPolygonX((XPDATA*)currentModel->pol[10], light);
            slPutPolygonX((XPDATA*)currentModel->pol[11], light);
            slPutPolygonX((XPDATA*)currentModel->pol[12], light);
            slPutPolygonX((XPDATA*)currentModel->pol[13], light);
            slPutPolygonX((XPDATA*)currentModel->pol[14], light);
            slPutPolygonX((XPDATA*)currentModel->pol[15], light);
        }
  }
  slPopMatrix();

    slPutPolygonX((XPDATA*)currentModel->pol[6], light);
    slPutPolygonX((XPDATA*)currentModel->pol[7], light);
    slPutPolygonX((XPDATA*)currentModel->pol[8], light);
    slPutPolygonX((XPDATA*)currentModel->pol[9], light);



}



ANGLE yy;
FIXED zDist=toFIXED(60.0);
Uint32 nbModels = 2;

#include "SONICR.h"
void			    my_draw(void)
{
    bool rtg = enableRTG;

    slPushMatrix();
    {
        computeLight();
        slTranslate(0, toFIXED(20.0), zDist);
        slRotX(DEGtoANG(6.0));
        slPushMatrix();
        {
            slRotY(yy);
            display_animated_model(&AnimCtrlPlay1, &entities[0], rtg);
        }
        slPopMatrix();

	 if (nbModels >= 2)
	 {
        slPushMatrix();
        {
            slRotY(yy);slTranslate(0, 0, -toFIXED(40.0));
            AnimCtrlPlay2.currentFrm++; if (AnimCtrlPlay2.currentFrm>15)AnimCtrlPlay2.currentFrm=0;
            animate_model(&entities[1], 0, AnimCtrlPlay2.currentFrm);
            //display_model(&entities[0], rtg);
        }
        slPopMatrix();
	  }

	 if (nbModels >= 3)
	 {

        slPushMatrix();
        {
            slRotY(yy);slTranslate(0, 0, toFIXED(40.0));
            display_model(&entities[0], 0);
        }
        slPopMatrix();
	 }

	 if (nbModels >= 4)
	 {

        slPushMatrix();
        {
            slRotY(yy);slTranslate(toFIXED(40.0), 0, 0);
            display_model(&entities[1], rtg);
        }
        slPopMatrix();

	}

	if (nbModels >= 5)
	 {

        slPushMatrix();
        {
            slRotY(yy);slTranslate(-toFIXED(40.0), 0, 0);
            display_model(&entities[1], 0);
        }
        slPopMatrix();

	}
	if (nbModels >= 6)
	 {

        slPushMatrix();
        {
            slRotY(yy);slTranslate(-toFIXED(40.0), 0, -toFIXED(40.0));
            display_model(&entities[1], 0);
        }
        slPopMatrix();

	}

	if (nbModels >= 7)
	 {

        slPushMatrix();
        {
            slRotY(yy);slTranslate(-toFIXED(40.0), 0, toFIXED(40.0));
            display_model(&entities[0], 0);
        }
        slPopMatrix();

	}


    }
slPopMatrix();


    slPrintFX(toFIXED(jo_3d_get_polygon_count()), slLocate(0,20));
    slPrintFX(toFIXED(jo_3d_get_vertices_count()), slLocate(0,21));
    slPrintFX(toFIXED(jo_3d_get_displayed_polygon_count()), slLocate(0,22));
	slPrintFX(toFIXED(framerate), slLocate(0,23));

    return;
}

void my_vblank()
{
    if(enableRTG == 1)
        slGouraudTblCopy();
}

void			my_gamepad(void)
{
    if (jo_is_pad1_key_down(PER_DGT_ST))
        {
         if (enableRTG == 1) enableRTG=0;
         else enableRTG=1;
        }
    if (jo_is_pad1_key_down(PER_DGT_TA))
        {
         nbModels++;
		if (nbModels>7) nbModels=1;
        }

    if (jo_is_pad1_key_pressed(PER_DGT_KU))
     zDist -= toFIXED(1.0);
    else if (jo_is_pad1_key_pressed(PER_DGT_KD))
	zDist += toFIXED(1.0);
    if (jo_is_pad1_key_pressed(PER_DGT_KR))
     yy+=(DEGtoANG(1.0));
    else if (jo_is_pad1_key_pressed(PER_DGT_KL))
	yy-=(DEGtoANG(1.0));

    if (jo_is_pad1_key_pressed(PER_DGT_TL))
	{SynchConst=(Sint8)1; framerate=1;}

    else if (jo_is_pad1_key_pressed(PER_DGT_TR))
	{SynchConst=(Sint8)2; framerate=2;}

	if (jo_is_pad1_key_down(PER_DGT_TC)) slDynamicFrame(ON);
	if (jo_is_pad1_key_down(PER_DGT_TZ)) slDynamicFrame(OFF);
}

void gameLoop(void)
{
   	while (1)
    {
        slUnitMatrix(0);
        my_gamepad();
        my_draw();
        slSynch();
    }
}

void			jo_main(void)
{
	jo_core_init(JO_COLOR_Black);

	/**Added by XL2 : turns off the TV while we load data**/
    slTVOff();
    /****/

    /**Added by XL2 **/
	//slDynamicFrame(ON); //Dynamic framerate, when the VDP1 can't fully draw to the framebuffer in the allocated amount of time (1/60, 2/60, etc.) it will continue drawing when it's ON. Else it will try to complete drawing by skipping lines and details and finish in the allocated time. If the app runs well with the value at OFF, leave it at OFF!
    SynchConst=(Sint8)2;  //Framerate control. 1/60 = 60 FPS, 2/60 = 30 FPS, etc.
	framerate=2;
	slZdspLevel(5);
    /****/



	/**XL2**/
	ztCDinit(); /**Reinit the CD functions using my own. Might break Jo Engine loading functions*/
    ztFont2NBG3();  /**Swaps the NBG0 to NBG3 (NBG0 is much more flexible, so we use NBG3 for text only)**/
    memset_w((Sint16*)(VDP2_VRAM_A0), 0x0000, 262144);
    ztLoadVDP2bmp("NBG0.ZTI", (Uint16*)(VDP2_VRAM_A1), 1, bmNBG0);  /** Just to show you how to use the NBG0 with my custom function **/
    fadeIn();

    /**Leave some time to show the splash screen**/
    /*int i=0;
    while (i<60*4){
        i+=2; slSynch();
    }*/

    fadeOut(1); /**Again just to show you, it clears the BG when you put (1), else it keeps it (0) **/
    ztLoadVDP2bmp("BEACH.ZTI", (Uint16*)(VDP2_VRAM_A0), 3, bmNBG1); /**Loads the Beach to NBG1**/
    slScrPosNbg1(toFIXED(0.0), toFIXED(50.0)); //Offset the background here

    /**ZTP stands for Z-Treme polygonal model**/
    void * currentAddress = (void*)LWRAM;

    gouraudCounter=0;
    currentAddress = ztLoad3Dmodel((Sint8*)"SONIC.ZTP", currentAddress, &entities[0], 1);
    currentAddress = ztLoad3Dmodel((Sint8*)"TAILS.ZTP", currentAddress, &entities[1], 1);
    //currentAddress = ztLoad3Dmodel((Sint8*)"TAILS.ZTP", currentAddress, &entities[2], 0);

    slInitGouraud(gourRealMax, GOUR_REAL_MAX, GRaddr, vwork);
	slIntFunction(my_vblank);
	/**Set your color here if you need one (depending on your light source)**/
    slSetGouraudColor(CD_White);

    /**To set the Jo Engine callback (might affect gouraud shading calculations)**/
    /*jo_core_add_callback(my_gamepad);
	jo_core_add_callback(my_draw);*/


	/**Animation control example : starts at frame 1 and ends at max frames - 2**/
	AnimCtrlPlay1.currentFrm=0;
	AnimCtrlPlay1.currentKeyFrm=0;
	AnimCtrlPlay1.endFrm=entities[0].nbFrames>>entities[0].AnimInterpolation;
	AnimCtrlPlay1.startFrm=1;

    AnimCtrlPlay2.currentFrm=0;
	AnimCtrlPlay2.currentKeyFrm=0;
	AnimCtrlPlay2.endFrm=0;
	AnimCtrlPlay2.startFrm=0;

    slScrAutoDisp(NBG0ON | NBG1ON | NBG3ON | SPRON);

	fadeIn();/**Smooth background fade in**/

	//jo_core_run();  //Uncomment for Jo Engine callback functions
	gameLoop();


}

/*
** END OF FILE
*/
